package gr.cite.annotation.integrationevent.inbox.annotationentitiestouch;

import gr.cite.annotation.audit.AuditableAction;
import gr.cite.annotation.common.JsonHandlingService;
import gr.cite.annotation.common.enums.IsActive;
import gr.cite.annotation.common.scope.fake.FakeRequestScope;
import gr.cite.annotation.common.scope.tenant.TenantScope;
import gr.cite.annotation.data.EntityUserEntity;
import gr.cite.annotation.data.TenantEntity;
import gr.cite.annotation.data.TenantEntityManager;
import gr.cite.annotation.integrationevent.inbox.EventProcessingStatus;
import gr.cite.annotation.integrationevent.inbox.InboxPrincipal;
import gr.cite.annotation.integrationevent.inbox.IntegrationEventProperties;
import gr.cite.annotation.model.Tenant;
import gr.cite.annotation.query.EntityUserQuery;
import gr.cite.annotation.query.TenantQuery;
import gr.cite.commons.web.oidc.principal.CurrentPrincipalResolver;
import gr.cite.commons.web.oidc.principal.extractor.ClaimExtractorProperties;
import gr.cite.tools.auditing.AuditService;
import gr.cite.tools.data.deleter.DeleterFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.logging.LoggerService;
import gr.cite.tools.validation.ValidatorFactory;
import jakarta.persistence.EntityManager;
import jakarta.persistence.EntityManagerFactory;
import jakarta.persistence.EntityTransaction;
import jakarta.persistence.OptimisticLockException;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class AnnotationEntitiesTouchedIntegrationEventHandlerImpl implements AnnotationEntitiesTouchedIntegrationEventHandler {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(AnnotationEntitiesTouchedIntegrationEventHandlerImpl.class));

    private final JsonHandlingService jsonHandlingService;

    private final ValidatorFactory validatorFactory;

    private final ApplicationContext applicationContext;

    private final QueryFactory queryFactory;

    public AnnotationEntitiesTouchedIntegrationEventHandlerImpl(JsonHandlingService jsonHandlingService, ValidatorFactory validatorFactory, ApplicationContext applicationContext, QueryFactory queryFactory) {
        this.jsonHandlingService = jsonHandlingService;
        this.validatorFactory = validatorFactory;
        this.applicationContext = applicationContext;
        this.queryFactory = queryFactory;
    }

    @Override
    public EventProcessingStatus handle(IntegrationEventProperties properties, String message) {
        AnnotationEntitiesTouchedIntegrationEvent event = this.jsonHandlingService.fromJsonSafe(AnnotationEntitiesTouchedIntegrationEvent.class, message);
        if (event == null)
            return EventProcessingStatus.Error;

        logger.debug("Handling {}", AnnotationEntitiesTouchedIntegrationEvent.class.getSimpleName());

        this.validatorFactory.validator(AnnotationEntitiesTouchedIntegrationEvent.AnnotationEntitiesTouchedIntegrationEventValidator.class).validateForce(event);

        EntityManager entityManager = null;
        EntityTransaction transaction = null;
        try (FakeRequestScope ignored = new FakeRequestScope()) {
            try {
                TenantScope scope = this.applicationContext.getBean(TenantScope.class);
                if (scope.isMultitenant() && properties.getTenantId() != null) {
                    TenantEntity tenant = queryFactory.query(TenantQuery.class).ids(properties.getTenantId()).firstAs(new BaseFieldSet().ensure(Tenant._id).ensure(Tenant._code));
                    if (tenant == null) {
                        logger.error("missing tenant from event message");
                        return EventProcessingStatus.Error;
                    }
                    scope.setTenant(properties.getTenantId(), tenant.getCode());
                } else if (scope.isMultitenant()) {
//                    logger.error("missing tenant from event message");
//                    return EventProcessingStatus.Error;
                    scope.setTenant(null, scope.getDefaultTenantCode());
                }

                CurrentPrincipalResolver currentPrincipalResolver = this.applicationContext.getBean(CurrentPrincipalResolver.class);

                ClaimExtractorProperties claimExtractorProperties = this.applicationContext.getBean(ClaimExtractorProperties.class);
                currentPrincipalResolver.push(InboxPrincipal.build(properties, claimExtractorProperties));

                EntityManagerFactory entityManagerFactory = this.applicationContext.getBean(EntityManagerFactory.class);
                entityManager = entityManagerFactory.createEntityManager();

                DeleterFactory deleterFactory = this.applicationContext.getBean(DeleterFactory.class);

                transaction = entityManager.getTransaction();
                transaction.begin();

                try {
                    TenantEntityManager tenantEntityManager = this.applicationContext.getBean(TenantEntityManager.class);
                    tenantEntityManager.setEntityManager(entityManager);
                    tenantEntityManager.disableTenantFilters();
                    for (AnnotationEntitiesTouchedIntegrationEvent.AnnotationEntityTouchedIntegrationEvent entityEvent : event.getEvents()) {

                        EntityUserQuery entityUserQuery = this.queryFactory.query(EntityUserQuery.class);
                        List<EntityUserEntity> items = entityUserQuery
                                .entityIds(entityEvent.getEntityId())
                                .isActive(IsActive.Active)
                                .collect();
                        List<UUID> updatedCreatedIds = new ArrayList<>();
                        for (UUID user : entityEvent.getUserIds()) {
                            EntityUserEntity data = items.stream().filter(x -> x.getUserId().equals(user)).findFirst().orElse(null);
                            if (data == null) {
                                data = new EntityUserEntity();
                                data.setId(UUID.randomUUID());
                                data.setEntityId(entityEvent.getEntityId());
                                data.setUserId(user);
                                data.setTenantId(properties.getTenantId());
                                data.setCreatedAt(Instant.now());
                                data.setUpdatedAt(Instant.now());
                                data.setIsActive(IsActive.Active);

                                entityManager.persist(data);
                            }
                            updatedCreatedIds.add(data.getId());
                        }

                        List<EntityUserEntity> toDelete = items.stream().filter(x -> updatedCreatedIds.stream().noneMatch(y -> y.equals(x.getId()))).collect(Collectors.toList());
                        deleterFactory.deleter(gr.cite.EntityUser.model.deleter.EntityUserDeleter.class).delete(toDelete);

                        entityManager.flush();
                    }

                    AuditService auditService = this.applicationContext.getBean(AuditService.class);

                    auditService.track(AuditableAction.User_Persist, Map.ofEntries(
                            new AbstractMap.SimpleEntry<String, Object>("model", event)
                    ));

                    transaction.commit();
                } catch (Exception e) {
                    transaction.rollback();
                    throw e;
                } finally {
                    currentPrincipalResolver.pop();
                }
            } catch (OptimisticLockException ex) {
                // we get this if/when someone else already modified the notifications. We want to essentially ignore this, and keep working
                logger.debug("Concurrency exception getting queue outbox. Skipping: {} ", ex.getMessage());
                if (transaction != null)
                    transaction.rollback();
            } catch (Exception ex) {
                logger.error("Problem getting list of queue outbox. Skipping: {}", ex.getMessage(), ex);
                if (transaction != null)
                    transaction.rollback();
            } finally {
                if (entityManager != null)
                    entityManager.close();
            }
        } catch (Exception ex) {
            logger.error("Problem getting list of queue outbox. Skipping: {}", ex.getMessage(), ex);
        }
        return EventProcessingStatus.Success;
    }

}
